#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#############################################################################
##                                                                         ##
## This file is part of DPAPIck                                            ##
## Windows DPAPI decryption & forensic toolkit                             ##
##                                                                         ##
##                                                                         ##
## Copyright (C) 2010, 2011 Cassidian SAS. All rights reserved.            ##
## Copyright (C) 2020       Howest. All rights reserved.                   ##
##                                                                         ##
##  Author:  Jean-Michel Picod <jmichel.p@gmail.com>                       ##
##  Updated: Photubias <tijl.deneut@howest.be>                             ##
##                                                                         ##
## This program is distributed under GPLv3 licence (see LICENCE.txt)       ##
##                                                                         ##
#############################################################################

import struct

class Eater(object):
    """This class is a helper for parsing binary structures."""

    def __init__(self, raw, offset=0, end=None, endianness="<"):
        self.raw = raw
        self.ofs = offset
        if end is None:
            end = len(raw)
        self.end = end
        self.endianness = endianness

    def prepare_fmt(self, fmt):
        """Internal use. Prepend endianness to the given format if it is not
        already specified.

        fmt is a format string for struct.unpack()

        Returns a tuple of the format string and the corresponding data size.

        """
        if fmt[0] not in ("<", ">", "!", "@"):
            fmt = self.endianness + fmt
        return fmt, struct.calcsize(fmt)

    def read(self, fmt):
        """Parses data with the given format string without taking away bytes.
        
        Returns an array of elements or just one element depending on fmt.

        """
        fmt, sz = self.prepare_fmt(fmt)
        v = struct.unpack_from(fmt, self.raw, self.ofs)
        if len(v) == 1:
            v = v[0]
        return v

    def eat(self, fmt):
        """Parses data with the given format string.
        
        Returns an array of elements or just one element depending on fmt.

        """
        fmt, sz = self.prepare_fmt(fmt)
        v = struct.unpack_from(fmt, self.raw, self.ofs)
        if len(v) == 1:
            v = v[0]
        self.ofs += sz
        return v

    def eat_string(self, length):
        """Eats and returns a string of length characters"""
        return self.eat("%us" % length)

    def eat_length_and_string(self, fmt):
        """Eats and returns a string which length is obtained after eating
        an integer represented by fmt

        """
        l = self.eat(fmt)
        return self.eat_string(l)

    def pop(self, fmt):
        """Eats a structure represented by fmt from the end of raw data"""
        fmt, sz = self.prepare_fmt(fmt)
        self.end -= sz
        v = struct.unpack_from(fmt, self.raw, self.end)
        if len(v) == 1:
            v = v[0]
        return v

    def pop_string(self, length):
        """Pops and returns a string of length characters"""
        return self.pop("%us" % length)

    def pop_length_and_string(self, fmt):
        """Pops and returns a string which length is obtained after poping an
        integer represented by fmt.

        """
        l = self.pop(fmt)
        return self.pop_string(l)
    
    def remain(self):
        """Returns all the bytes that have not been eated nor poped yet."""
        return self.raw[self.ofs:self.end]

    def eat_sub(self, length):
        """Eats a sub-structure that is contained in the next length bytes"""
        sub = self.__class__(self.raw[self.ofs:self.ofs+length], endianness=self.endianness)
        self.ofs += length
        return sub

    #def __nonzero__(self):
    def __bool__(self):
        return self.ofs < self.end


class DataStruct(object):
    """Don't use this class unless you know what you are doing!"""

    def __init__(self, raw=None):
        if raw is not None:
            self.parse(Eater(raw, endianness="<"))

    def parse(self, eater_obj):
        raise NotImplementedError("This function must be implemented in subclasses")




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#############################################################################
##                                                                         ##
## This file is part of DPAPIck                                            ##
## Windows DPAPI decryption & forensic toolkit                             ##
##                                                                         ##
##                                                                         ##
## Copyright (C) 2010, 2011 Cassidian SAS. All rights reserved.            ##
## Copyright (C) 2020       Howest. All rights reserved.                   ##
##                                                                         ##
##  Author:  Jean-Michel Picod <jmichel.p@gmail.com>                       ##
##  Updated: Photubias <tijl.deneut@howest.be>                             ##
##                                                                         ##
## This program is distributed under GPLv3 licence (see LICENCE.txt)       ##
##                                                                         ##
#############################################################################

from Registry import Registry
import dpapick3.crypto as crypto
import dpapick3.eater as eater


class Regedit(object):
    """This class provides several functions to handle registry extraction stuff.
    """

    def __init__(self):
        self.syskey = None
        self.lsakeys = None
        self.policy = {"major": 0, "minor": 0, "value": 0}
        self.lsa_secrets = {}

    def get_syskey(self, system):
        """Returns the syskey value after decryption from the registry values.
        system argument is the full path to the SYSTEM registry file (usually
        located under %WINDIR%\\system32\\config\\ directory.

        """
        with open(system, 'rb') as f:
            r = Registry.Registry(f)
            cs = r.open('Select').value('Current').value()
            r2 = r.open("ControlSet%03d\\Control\\Lsa" % cs)
            syskey = ''.join([r2.subkey(x)._nkrecord.classname() for x in ('JD', 'Skew1', 'GBG', 'Data')])
        syskey = bytes.fromhex(syskey)
        transforms = [8, 5, 4, 2, 11, 9, 13, 3, 0, 6, 1, 12, 14, 10, 15, 7]
        self.syskey = ''
        for i in range(len(syskey)):
            self.syskey += '{:02x}'.format(syskey[transforms[i]])
        self.syskey = bytes.fromhex(self.syskey)
        return self.syskey

    def get_lsa_key(self, security):
        """Returns and decrypts the LSA secret key for "CurrentControlSet".
        It is stored under Policy\\PolSecretEncryptionKey.

        security is the full path the the SECURITY registry file (usually
        located under %WINDIR%\\system32\\config\\ directory.

        To decrypt the LSA key, syskey is required. Thus you must first call
        self.get_syskey() if it has not been previously done.
        """
        lsakey = ''
        if self.syskey is None:
            raise ValueError('Must provide syskey or call get_syskey() method first')
        with open(security, 'rb') as f:
            r = Registry.Registry(f)
            rev = eater.Eater(r.open("Policy\\PolRevision").value("(default)").value())
            self.policy["minor"] = rev.eat("H")
            self.policy["major"] = rev.eat("H")
            self.policy["value"] = float("%d.%02d" % (self.policy["major"], self.policy["minor"]))
            if self.policy["value"] > 1.09:
                # NT6
                r2 = r.open("Policy\\PolEKList")
                lsakey = r2.value("(default)").value()
            else:
                # NT5
                r2 = r.open("Policy\\PolSecretEncryptionKey")
                lsakey = r2.value("(default)").value()
        rv = None
        if self.policy['value'] > 1.09:
            currentKey, self.lsakeys = crypto.decrypt_lsa_key_nt6(lsakey, self.syskey)
            rv = self.lsakeys[currentKey]['key']
        else:
            self.lsakeys = crypto.decrypt_lsa_key_nt5(lsakey, self.syskey)
            rv = self.lsakeys[1]
        return rv

    def get_lsa_secrets(self, security, system):
        """Retrieves and decrypts LSA secrets from the registry.
        security and system arguments are the full path to the corresponding
        registry files.
        This function automatically calls self.get_syskey() and
        self.get_lsa_key() functions prior to the secrets retrieval.

        Returns a dictionary of secrets.

        """
        self.get_syskey(system)
        currentKey = self.get_lsa_key(security)
        self.lsa_secrets = {}
        with open(security, 'rb') as f:
            r = Registry.Registry(f)
            r2 = r.open('Policy\\Secrets')
            for i in r2.subkeys():
                self.lsa_secrets[i.name()] = {}
                for j in i.subkeys():
                    self.lsa_secrets[i.name()][j.name()] = j.value('(default)').value()
        for k, v in list(self.lsa_secrets.items()):
            for s in ['CurrVal', 'OldVal']:
                if v[s] != b'':
                    if self.policy['value'] > 1.09:
                        # NT6
                        self.lsa_secrets[k][s] = crypto.decrypt_lsa_secret(v[s], self.lsakeys)
                    else:
                        self.lsa_secrets[k][s] = crypto.SystemFunction005(v[s][0xc:], currentKey)
            for s in ['OupdTime', 'CupdTime']:
                #print(int(self.lsa_secrets[k][s].hex(),16))
                #if self.lsa_secrets[k][s] > 0:
                if not self.lsa_secrets[k][s] == b'':
                    t = eater.Eater(self.lsa_secrets[k][s])
                    self.lsa_secrets[k][s] = int((t.eat('Q') / 10000000) - 11644473600)
        #print(self.lsa_secrets)
        return self.lsa_secrets
    
    def getUsername(self, software, sid):
        """Retrieves username from the SOFTWARE registry when given the full user SID
        """
        with open(software, 'rb') as oFile:
            oReg = Registry.Registry(oFile)
            oRegKey = oReg.open('Microsoft\\Windows\\CurrentVersion\\Authentication\\Credential Providers\{D6886603-9D2F-4EB2-B667-1971041FA96B}')
            for oKey in oRegKey.subkeys():
                if sid in oKey.name():
                    return oKey.subkey('UserNames').subkeys()[0].name()
        return '<Unknown>'
    '''
    def getUsername(self, sam, rid):
        """Retrieves username from the SAM registry when given the user RID (integer)
        """
        with open(sam, 'rb') as oFile:
            oReg = Registry.Registry(oFile)
            oRegKey = oReg.open("SAM\\Domains\\Account\\Users\\Names")
            for oKey in oRegKey.subkeys():
                if int(oKey.value('(default)').value_type()) == rid: return oKey.name()
        return '<Unknown>'
    '''
# vim:ts=4:expandtab:sw=4




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#############################################################################
##                                                                         ##
## This file is part of DPAPIck                                            ##
## Windows DPAPI decryption & forensic toolkit                             ##
##                                                                         ##
##                                                                         ##
## Copyright (C) 2010, 2011 Cassidian SAS. All rights reserved.            ##
## Copyright (C) 2020       Howest. All rights reserved.                   ##
##                                                                         ##
##  Author:  Jean-Michel Picod <jmichel.p@gmail.com>                       ##
##  Updated: Photubias <tijl.deneut@howest.be>                             ##
##                                                                         ##
## This program is distributed under GPLv3 licence (see LICENCE.txt)       ##
##                                                                         ##
#############################################################################

import struct
import hashlib
import dpapick3.crypto as crypto
import dpapick3.eater as eater

class RPC_SID(eater.DataStruct):
    """Represents a RPC_SID structure. See MSDN for documentation"""
    def __init__(self, raw=None):
        self.version = None
        self.idAuth = None
        self.subAuth = None
        eater.DataStruct.__init__(self, raw)

    def parse(self, data):
        self.version = data.eat("B")
        n = data.eat("B")
        self.idAuth = struct.unpack(">Q", b'\0\0' + data.eat("6s"))[0]
        self.subAuth = data.eat("%dL" % n)

    def __str__(self):
        s = ["S-%d-%d" % (self.version, self.idAuth)]
        s += ["%d" % x for x in self.subAuth]
        return "-".join(s)

    def __repr__(self):
        return """RPC_SID(%s):
        revision             = %d
        identifier-authority = %r
        subAuthorities       = %r""" % (self, self.version, self.idAuth, self.subAuth)


class CredSystem(eater.DataStruct):
    """This represents the DPAPI_SYSTEM token which is stored as an LSA
        secret.

        Sets 2 properties:
            self.machine
            self.user

    """
    def __init__(self, raw=None):
        self.machine = None
        self.user = None
        self.revision = None
        eater.DataStruct.__init__(self, raw)

    def parse(self, data):
        self.revision = data.eat("L")
        self.machine = data.eat("20s")
        self.user = data.eat("20s")

    def __repr__(self):
        s = ["DPAPI_SYSTEM:"]
        if self.user is not None:
            s.append("\tUser Credential   : %s" % self.user.hex())
        if self.machine is not None:
            s.append("\tMachine Credential: %s" % self.machine.hex())
        return "\n".join(s)


class CredhistEntry(eater.DataStruct):
    """Represents an entry in the Credhist file"""
    def __init__(self, raw=None):
        self.pwdhash = None
        self.hmac = None
        self.revision = None
        self.hashAlgo = None
        self.rounds = None
        self.cipherAlgo = None
        self.shaHashLen = None
        self.ntHashLen = None
        self.iv = None
        self.userSID = None
        self.encrypted = None
        self.revision2 = None
        self.guid = None
        self.ntlm = None
        eater.DataStruct.__init__(self, raw)

    def __getstate__(self):
        d = dict(self.__dict__)
        for k in ["cipherAlgo", "hashAlgo"]:
            if k in d:
                d[k] = d[k].algnum
        return d

    def __setstate__(self, d):
        for k in ["cipherAlgo", "hashAlgo"]:
            if k in d:
                d[k] = crypto.CryptoAlgo(d[k])
        self.__dict__.update(d)

    def parse(self, data):
        self.revision = data.eat("L")
        self.hashAlgo = crypto.CryptoAlgo(data.eat("L"))
        self.rounds = data.eat("L")
        data.eat("L")
        self.cipherAlgo = crypto.CryptoAlgo(data.eat("L"))
        self.shaHashLen = data.eat("L")
        self.ntHashLen = data.eat("L")
        self.iv = data.eat("16s")

        self.userSID = RPC_SID()
        self.userSID.parse(data)

        n = self.shaHashLen + self.ntHashLen
        n += -n % self.cipherAlgo.blockSize
        self.encrypted = data.eat_string(n)

        self.revision2 = data.eat("L")
        self.guid = "%0x-%0x-%0x-%0x%0x-%0x%0x%0x%0x%0x%0x" % data.eat("L2H8B")

    def decryptWithKey(self, enckey):
        """Decrypts this credhist entry using the given encryption key."""
        cleartxt = crypto.dataDecrypt(self.cipherAlgo, self.hashAlgo, self.encrypted,
                                      enckey, self.iv, self.rounds)
        self.pwdhash = cleartxt[:self.shaHashLen]
        self.ntlm = cleartxt[self.shaHashLen:self.shaHashLen + self.ntHashLen].rstrip(b'\x00')
        if len(self.ntlm) != 16:
            self.ntlm = None

    def decryptWithHash(self, pwdhash):
        """Decrypts this credhist entry with the given user's password hash.
        Simply computes the encryption key with the given hash then calls
        self.decryptWithKey() to finish the decryption.

        """
        self.decryptWithKey(crypto.derivePwdHash(pwdhash, str(self.userSID)))

    def decryptWithPassword(self, password):
        """Decrypts this credhist entry with the given user's password.
        Simply computes the password hash then calls self.decryptWithHash()

        """
        return self.decryptWithHash(hashlib.sha1(password.encode("UTF-16LE")).digest())

    def jtr_shadow(self):
        """Returns a string that can be passed to John the Ripper to crack this
            CREDHIST entry. Requires to use a recent jumbo version of JtR plus
            the configuration snipplet in the "3rdparty" directory of DPAPIck.

            Unless you know what you are doing, you shall not call this function
            yourself. Instead, use the method provided by CredHistPool object.
        """
        rv = []
        if self.pwdhash is not None:
            rv.append("%s:$dynamic_1400$%s" % (self.userSID, self.pwdhash.encode('hex')))
        if self.ntlm is not None:
            rv.append("%s:$NT$%s" % (self.userSID, self.ntlm.encode('hex')))
        return "\n".join(rv)

    def __repr__(self):
        s = ["CredHist entry",
             "\trevision   = %x\n" % self.revision,
             "\thash       = %r" % self.hashAlgo,
             "\trounds     = %i" % self.rounds,
             "\tcipher     = %r" % self.cipherAlgo,
             "\tshaHashLen = %i" % self.shaHashLen,
             "\tntHashLen  = %i" % self.ntHashLen,
             "\tuserSID    = %s" % self.userSID,
             "\tguid       = %s" % self.guid,
             "\tiv         = %s" % self.iv.hex()]
        if self.pwdhash is not None:
            s.append("\tpwdhash  = %s" % self.pwdhash.hex())
        if self.ntlm is not None:
            s.append("\tNTLM     = %s" % self.ntlm.hex())
        return "\n".join(s)


class CredHistFile(eater.DataStruct):
    """Represents a CREDHIST file.
    Be aware that currently, it is not possible to check whether the decryption
    succeeded or not. To circumvent that and optimize a little bit crypto
    operations, once a credhist entry successfully decrypts a masterkey, the
    whole CredHistFile is flagged as valid. Then, no further decryption occurs.

    """
    def __init__(self, raw=None):
        self.entries_list = []
        self.entries = {}
        self.valid = False
        self.footmagic = None
        self.curr_guid = None
        eater.DataStruct.__init__(self, raw)

    def parse(self, data):
        while True:
            l = data.pop("L")
            if l == 0:
                break
            self.addEntry(data.pop_string(l - 4))

        self.footmagic = data.eat("L")
        self.curr_guid = "%0x-%0x-%0x-%0x%0x-%0x%0x%0x%0x%0x%0x" % data.eat("L2H8B")

    def addEntry(self, blob):
        """Creates a CredhistEntry object with blob then adds it to the store"""
        x = CredhistEntry(blob)
        self.entries[x.guid] = x
        self.entries_list.append(x)

    def validate(self):
        """Simply flags a file as successfully decrypted. See the class
        documentation for information.

        """
        self.valid = True

    def decryptWithHash(self, h):
        """Try to decrypt each entry with the given hash"""
        if self.valid:
            return
        curhash = h
        for entry in self.entries_list:
            entry.decryptWithHash(curhash)
            curhash = entry.pwdhash

    def decryptWithPassword(self, pwd):
        """Try to decrypt each entry with the given password.
        This function simply computes the SHA-1 hash with the password, then
        calls self.decryptWithHash()

        """
        return self.decryptWithHash(hashlib.sha1(pwd.encode("UTF-16LE")).digest())

    def jtr_shadow(self, validonly=False):
        """Returns a string that can be passed to John the Ripper to crack the
            CREDHIST entries. Requires to use a recent jumbo version of JtR plus
            the configuration snipplet in the "3rdparty" directory of DPAPIck.

            If validonly is set to True, will only extract CREDHIST entries
            that are known to have sucessfully decrypted a masterkey.

        """
        if validonly and not self.valid:
            return ""
        s = []
        #for e in self.entries.itervalues():
        for e in list(self.entries.values()):
            s.append(e.jtr_shadow())
        return "\n".join(s)

    def __repr__(self):
        s = ["CredHistPool:  %s" % self.curr_guid]
        #for e in self.entries.itervalues():
        for e in list(self.entries.values()):
            s.append("---")
            s.append(repr(e))
        s.append("====")
        return "\n".join(s)


# vim:ts=4:expandtab:sw=4



#!/usr/bin/env python3
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="dpapick3",
    version="0.1.5",
    author="Tijl Deneut",
    author_email="tijl.deneut@howest.be",
    description="A native implementation of DPAPI",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/tijldeneut/dpapick3",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.2',
    install_requires=[
        'pycryptodome',
        'python-registry'
    ]
)